unit UAi;

interface
Uses UInteractor;

Type
  TFuncOutput = Function (Str : String) : Integer of object;

  PMoveNode = ^MoveNode;
  MoveNode = Record
    x : Byte;
    y : Byte;
    rightClick : boolean;
    next       : PMoveNode;
  end;

TMoveQueue = class
public
  numMoves : SmallInt;
  Constructor Create;
  Destructor Destroy;
  Function Enqueue(x,y : Byte;rightClick : Boolean = false) : Boolean;
  Function Dequeue(var x,y : Byte;var rightClick : Boolean) : Boolean;
  Function IsEmpty() : Boolean;
  Procedure EmptyList();
  Function GetIndex(index : SmallInt): PMoveNode ;
private
  first : PMoveNode ;
  last  : PMoveNode ;
  output : TFuncOutput;
end;

Const
  MSAI_ERROR			='E';
  MSAI_SUCCESS			=1;
  MSAI_MATRIXSIZE2		=3;
  MSAI_VERBOSE_LEVELS		=6;
  MSAI_VERBOSE_DEFAULTS		=$13;
  MSAI_MOVE_NONEFOUND		=0;
  MSAI_MOVE_UNITPROPAGATION	=1;
  MSAI_MOVE_CONTRADICTION	=2;
  MSAI_MOVE_PROBABILITY	        =3;

Type
  TMinesweeperAI = class
  public
    ////////////////////////////////////////////////////////////////////////////////////////
    // Public Member Functions
    ////////////////////////////////////////////////////////////////////////////////////////
   verbose : array[0..MSAI_VERBOSE_LEVELS-1] of boolean;
   stopOnProbability : Boolean;
   msi : TDemineurInteractor;
   output : TFuncOutput;
   Constructor Create;
   Destructor Destroy;override;

   Function CreateBoard(): Byte;
   Function GetPlayState() : byte;
   Procedure OutputState();
   Function GetMove() : Byte;
   Function SolveBoard(Var numUnitPropagation,numContradiction,numProbability : SmallInt) : Byte;
   Function ExecuteQueue() : Byte;

private
  ////////////////////////////////////////////////////////////////////////////////////////
  // Private Member Functions
  ////////////////////////////////////////////////////////////////////////////////////////
  board : Pchar;
  columns, rows : Byte;
  numMinesLeft, numBlanks : SmallInt;
  moveQueue : TMoveQueue ;
  h : array[0..(2*MSAI_MATRIXSIZE2+1)*(2*MSAI_MATRIXSIZE2+1)-1] of integer ;
  Function GetBoard(col,row : integer) : Char;
  Function SetBoard(col,row : Byte;symbol : char) : Byte;
  Function FindContradiction() : Byte;
  Function GetMoveProbability(): Byte;
  Function GetMatrix(col,row : byte): Byte;
  Function ReduceSatisfied(col,row : Byte): Byte;
  Function CheckUnsatisfiable(col,row : Byte): Byte;
  Function Map2Dto1D(column,row : Byte) : SmallInt;
  Function GetUnitPropagation(col,row : Byte) : SmallInt;
  Function GetNear(col,row : Byte;Var blanks,flags : Byte) : byte;
  Function GetNearUncovered(col,row : byte) : Byte;
  Function QueueSurrounding(col,row : Byte;flag : Boolean) : Byte;
  Function GetProbability(col,row : Byte) : double ;
end;

implementation
Uses SysUtils,Windows;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Constructor TMoveQueue.Create;
begin
  first := NIL;
  last := NIL;
  numMoves := 0;
end;

Destructor TMoveQueue.Destroy;
begin
  EmptyList();
end;

Procedure TMoveQueue.EmptyList();
var
  temp : PMoveNode ;
begin
  while(first <> NIL) do
  begin
    temp := first.next;
    Dispose(first);
    first := temp;
  end;
  first := NiL;
  last  := NiL;
  numMoves := 0;
end;

Function TMoveQueue.IsEmpty() : boolean;
begin
  Result := first = NiL;
end;

Function TMoveQueue.Enqueue(x,y : Byte;rightClick : Boolean) : Boolean;
Var
  current,
  newNode : PMoveNode;
begin
  current := first;
  while(current <> niL) do
  begin
    if((current.x = x) and (current.y = y)) Then
    begin
      Result := false;
      Exit;
    end;
    current := current.next;
  end;

  newNode := new(PMoveNode);
  newNode.x := x;
  newNode.y := y;
  newNode.next := NiL;
  newNode.rightClick := rightClick;

  if(first = NiL) Then
  begin
    first := newNode;
    last  := newNode ;
  end
  else
  begin
    last.next := newNode;
    last := newNode;
  end;
  Inc(numMoves);
  Result := true;
end;

Function TMoveQueue.Dequeue(Var x,y : Byte;Var rightClick : Boolean) : Boolean;
var
  temp : PMoveNode;
begin
  Result := false;
  if(first <> NIL) Then
  Begin
    x := first.x;
    y := first.y;
    rightClick := first.rightClick;

    temp := first.next;
    Dispose(first);
    first := temp;

    if(first = NiL) Then
      last := NiL;

    dec(numMoves);
    Result := true;
  end;
end;

Function TMoveQueue.GetIndex(index : SmallInt) : PMoveNode;
Var
  i : SmallInt;
  current : PMoveNode;
begin
  i := 0;
  current := first;
  while ((i < index) and (current <> Nil)) do
  begin
    Inc(i);
    current := current.next;
  end;
  Result := current;
end;

Constructor TMinesweeperAI.Create ;
var
  i : Byte;
begin
  board        := Nil;
  numMinesLeft := 0;
  numBlanks    := 0;
  moveQueue := TmoveQueue.Create;
  msi := TDemineurInteractor.Create;
  stopOnProbability := false;
  for i:=0 to  pred(MSAI_VERBOSE_LEVELS) do
   verbose[i] := (MSAI_VERBOSE_DEFAULTS and (1 shl i)) <> 0;
end;

Destructor TMinesweeperAI.Destroy;
Begin
  if assigned(board) Then
    Dispose(board);
   msi.Free;
end;

Function TMinesweeperAI.GetBoard(col,row : integer) : Char;
begin
  if((col >= 0) and (col < columns) and (row >= 0) and (row < rows)) Then
    Result := board[Map2Dto1D(col, row)]
  else
    Result := MSAI_ERROR;
end;

Function TMinesweeperAI.SetBoard(col,row : Byte;symbol : char) : Byte;
Begin
  if ( assigned(board)
    and (col >= 0)
    and (col <= columns)
    and (row >= 0)
    and (row <= rows)) Then
  begin
    board[Map2Dto1D(col, row)] := symbol;
    Result := MSAI_SUCCESS;
  end
  else
    Result := ord(MSAI_ERROR);
end;

Function TMinesweeperAI.CreateBoard() : Byte;
var
  oldColumns, oldRows : Byte;
Begin
  oldColumns := columns;
  oldRows    := rows;

  if(msi.GetPlayAreaInfo() = MSI_ERROR) Then
  begin
    Result := MSI_ERROR;
    Exit;
  end;
  columns := msi.GetColumns();
  rows := msi.GetRows();
  if((columns <> oldColumns) or (rows <> oldRows)) Then
  begin
    if(verbose[1]) Then
      output(Format('Creating new %u x %u board',[columns, rows]));

    if assigned(board) then
       FreeMem(board);
    GetMem(board,columns*rows);
  end;
  Result := MSAI_SUCCESS;
end;

Function TMinesweeperAI.GetPlayState() : Byte;
var
  symbol : Char;
  x,y: Byte;
  buffer : String[10];
begin
  numMinesLeft := msi.GetMinesLeft();
  numBlanks    := 0;
  x:= 0;
  y:= 0;
  CreateBoard();
  for y := 0 To pred(rows) do
  begin
    for x :=0 To pred(columns) do
    begin
      symbol := msi.GetSymbol(x,y);
      if(symbol = char(MSI_ERROR)) Then
	output(Format('MSI_ERROR: GetPlayState() cannot get symbol at %u, %u!',[x,y]));
      if(symbol = ' ') Then
	inc(numBlanks);
      SetBoard(x, y, symbol);
    end;
  end;

  if(verbose[1]) Then
  begin
    if(numBlanks = 0) Then
      buffer := format('game over',[])
    else
    buffer := format('%.1f%%',[numMinesLeft / numBlanks]);
    output(Format('GetPlayState(): %u mines left and %u blanks (%s)',[numMinesLeft, numBlanks,buffer]));
  end;
  Result := msi.GetGoalState();
end;

Procedure TMinesweeperAI.OutputState();
var
  buf : array[0..128] of Char;
  count : Integer;
  x,y : Byte;
begin
  if(verbose[0]) Then
  Begin
    count := 0;
    for y := 0 to rows do
    begin
      for x := 0 to columns  do
      begin
       	buf[count] := GetBoard(x, y);
        inc(count);
       	if((ord(buf[count-1]) >= 0) and (ord(buf[count-1]) <= 8)) Then
	  buf[count] := '0';
      end;
      inc(count);
      buf[count] := '0';
      count := 0;
      output(buf);
    end;
  end;
end;

Function iif(bool : boolean;aval1,aval2 : string): string;
begin
  if bool Then
    result := aval1
  else
    result := aval2;
end;

Function TMinesweeperAI.GetMove() :Byte;
var
   x, y : Byte;
begin
  if(verbose[1]) Then
    output('');
  GetPlayState();
  moveQueue.EmptyList();
  if(verbose[1]) Then
    output('Looking for unit propagation moves ...');
  for x := 0 To pred(columns) do
   for y := 0 To  pred(rows) do
      GetUnitPropagation(x, y);

   if(moveQueue.numMoves = 0) Then
   begin
     if(verbose[1]) Then
     begin
       output('     ... No possible unit propagation moves');
       output('Looking for contradictions ...');
     end;
     if(FindContradiction() = MSAI_MOVE_CONTRADICTION) Then
     begin
       Result := MSAI_MOVE_CONTRADICTION;// If no contradictions are found, resort to best-guess probability.
       Exit;
     end
     else
     begin
       if(verbose[1]) then
       begin
	 output('     ... no contradictions found.');
	 output('Picking a square based on probability ...');
       end;
       result := GetMoveProbability();
       Exit;
     end;
   end
   else		// Unit propagation has found optimal moves and has placed them in the queue.
   begin
     if(verbose[1]) then
       output(Format('     ... found %u move%s, placed %s in queue',[moveQueue.numMoves,iif(moveQueue.numMoves = 1,'','s'),iif(moveQueue.numMoves = 1,'it','them')]));
     Result := MSAI_MOVE_UNITPROPAGATION;
     Exit;
   end;
end;

Function TMinesweeperAI.GetUnitPropagation(col,row : Byte) : SmallInt;
var
  symbol : Char;
  numMoves : SmallInt;
  blanks, flags : Byte;
begin
  numMoves := 0;
  symbol := GetBoard(col, row);

  if((ord(symbol) >= 0) and (ord(symbol) <= 8)) Then
  begin
    GetNear(col,row, blanks,flags);
    if(flags = Ord(symbol)) Then
      numMoves := numMoves +QueueSurrounding(col, row, false)
    else if(ord(symbol) = blanks + flags) Then
      numMoves := numMoves +QueueSurrounding(col, row, true);
  end;
  Result := numMoves;
end;

Function TMinesweeperAI.FindContradiction() : Byte;
Var
  x, y : Byte;
  symbol : Char;
  uncovered : Byte;
  index : Byte;
  i,j : Byte;
begin
  for y := 0 to rows do
  begin
    for x := 0 to columns do
    begin
      symbol := GetBoard(x, y);
      uncovered := GetNearUncovered(x, y);
      if ((symbol = MSI_SQUARE_BLANK ) and (uncovered <> 0)) Then
      begin
	if(verbose[2]) Then
	  output(format('Square %u, %u is blank and has %u uncovered ...',[x, y, uncovered]));
	GetMatrix(x, y);
	if(verbose[3]) Then
	begin
	  output('After step 1 h[] = {');
	  for i := 0 to (2*MSAI_MATRIXSIZE2+1) do
	     output(format('          %4.0f  %4.0f  %4.0f  %4.0f  %4.0f  %4.0f  %4.0f',[h[i*(2*MSAI_MATRIXSIZE2+1)], h[i*(2*MSAI_MATRIXSIZE2+1)+1], h[i*(2*MSAI_MATRIXSIZE2+1)+2], h[i*(2*MSAI_MATRIXSIZE2+1)+3], h[i*(2*MSAI_MATRIXSIZE2+1)+4], h[i*(2*MSAI_MATRIXSIZE2+1)+5], h[i*(2*MSAI_MATRIXSIZE2+1)+6]]));
	end;
	ReduceSatisfied(x, y);
	if(verbose[3]) then
	begin
	  output('After step 2 h[] = {');
	  for i := 0 to (2*MSAI_MATRIXSIZE2+1) do
	    output(format('          %4.0f  %4.0f  %4.0f  %4.0f  %4.0f  %4.0f  %4.0f',[h[i*(2*MSAI_MATRIXSIZE2+1)], h[i*(2*MSAI_MATRIXSIZE2+1)+1], h[i*(2*MSAI_MATRIXSIZE2+1)+2], h[i*(2*MSAI_MATRIXSIZE2+1)+3], h[i*(2*MSAI_MATRIXSIZE2+1)+4], h[i*(2*MSAI_MATRIXSIZE2+1)+5], h[i*(2*MSAI_MATRIXSIZE2+1)+6]]));
	end;
	if(CheckUnsatisfiable(x, y) = MSAI_MOVE_CONTRADICTION) then
        begin
	  Result := MSAI_MOVE_CONTRADICTION;
          exit;
        end;
      end;
    end;
  end;
  Result := MSAI_MOVE_NONEFOUND;
end;

Function TMinesweeperAI.GetMatrix(col,row : byte) : Byte;
var
  i, j  : Smallint;
  index : Byte;
   symbol :Char;
begin
  for j := -MSAI_MATRIXSIZE2 to MSAI_MATRIXSIZE2 do
  begin
    for i := -MSAI_MATRIXSIZE2 to  MSAI_MATRIXSIZE2 do
    begin
      index  := (i+MSAI_MATRIXSIZE2)+(j+MSAI_MATRIXSIZE2)*(2*MSAI_MATRIXSIZE2+1);
      symbol := GetBoard(col+i, row+j);

      if(j = 0) and (i = 0) Then
       	h[index] := -1
      else if(symbol = MSI_SQUARE_FLAG)  Then
       	h[index] := -1
      else if(symbol = MSAI_ERROR) Then
       	h[index] := 0
      else if(symbol = MSI_SQUARE_BLANK) Then
	h[index] := -128
      else
	h[index] := ord(symbol);
    end;
  end;
  Result := MSAI_SUCCESS;
end;

Function TMinesweeperAI.ReduceSatisfied(col,row : Byte) : Byte;
var
  i, j : SmallInt;
  index : integer;
  symbol : Char;
  u,v : Shortint;
  blanks, flags : Byte;
  aval : integer;
begin
  for j := -MSAI_MATRIXSIZE2 to MSAI_MATRIXSIZE2 do
  begin
    for i := -MSAI_MATRIXSIZE2 to MSAI_MATRIXSIZE2 do
    begin
      index := (i+MSAI_MATRIXSIZE2)+(j+MSAI_MATRIXSIZE2)*(2*MSAI_MATRIXSIZE2+1);
      symbol := GetBoard(col + i, row + j);
      if ((ord(symbol) >= 1) and (ord(symbol) <= 8)) Then
      begin
	GetNear(col + i, row + j,blanks,flags);
	h[index] := h[index] - flags;
	if  ( ((i = -1) or (i = 1) or (i = 0))
        and   ((j = -1) or (j = 1) or (j = 0))
        and   ((i <> 0) or (j <> 0))) Then
        begin
          aval := h[index];
          dec(aval);
	  h[index] := aval;
        end;
	if(h[index] = 0) Then
	begin
	  for u := -1 to 1  do
	  begin
	    for v := -1 to 1 do
	    begin
	      if  ((-MSAI_MATRIXSIZE2 <= (i+u))
              and  ((i+u) <= MSAI_MATRIXSIZE2)
              and  (-MSAI_MATRIXSIZE2 <= (j+v))
              and  ((j+v) <= MSAI_MATRIXSIZE2)
              and  ((u <> 0) or (v <> 0))) Then
	      begin
		if(h[index+u+v*(2*MSAI_MATRIXSIZE2+1)] = -128) Then
	          h[index+u+v*(2*MSAI_MATRIXSIZE2+1)] := -2;
	      end;
	    end;
	  end;
	end;
      end;
    end;
  end;
  Result := MSAI_SUCCESS;
end;

Function TMinesweeperAI.CheckUnsatisfiable(col,row : Byte) : Byte;
Var
  validBlanks : byte ;
  i, j : smallInt;
  found : Boolean;
  index : Byte;
  symbol : Char;
  flags : Byte;
  u,v : Shortint;
begin
  moveQueue.EmptyList();
  found := false;
  for j := -MSAI_MATRIXSIZE2  to MSAI_MATRIXSIZE2 do
  begin
    for i := -MSAI_MATRIXSIZE2  To MSAI_MATRIXSIZE2 do
    begin
      index  := (i+MSAI_MATRIXSIZE2)+(j+MSAI_MATRIXSIZE2)*(2*MSAI_MATRIXSIZE2+1);
      symbol := GetBoard(col + i, row + j);
      if((ord(symbol) >= 1) and (ord(symbol) <= 8)) Then
      begin
	GetNear(col+i, row+j,validBlanks,flags);
	for u := -1 to 1  do
	  for v := -1 to 1 do
	  begin
	    if ((-MSAI_MATRIXSIZE2 <= (i+u))
              and ((i+u) <= MSAI_MATRIXSIZE2)
              and (-MSAI_MATRIXSIZE2 <= (j+v))
              and ((j+v) <= MSAI_MATRIXSIZE2)
              and ((u <> 0) or (v <> 0))) Then
	    begin
	      if(h[index+u+v*(2*MSAI_MATRIXSIZE2+1)] = -2) Then
                if validBlanks > 0 Then
	          dec(validBlanks);
	    end;
	  end;
	  if  (((i = -1) or (i = 1) or (i = 0))
          and ((j = -1)  or (j = 1) or (j = 0))
          and ((i <> 0)  or (j <> 0))) Then
           if validBlanks > 0 Then
             dec(validBlanks);
	  if(verbose[3]) Then
	    output(format('%u < h[%u]=%g ? ',[validBlanks, index, h[index]]));
	  if(validBlanks < h[index]) then
	  begin
	    if(verbose[1]) Then
	      output(Format('     ... Contradiction found!  Added move: %u,%u for uncovering to queue',[col, row]));
            moveQueue.Enqueue(col, row);
	    found := true;
	  end;
      end;
    end;
  end;
  if(found) Then
    Result := MSAI_MOVE_CONTRADICTION
  else
    Result := MSAI_MOVE_NONEFOUND;
end;


Function TMinesweeperAI.SolveBoard(var numUnitPropagation,numContradiction,numProbability : SmallInt) : Byte;
Var
  done : Boolean;
  lastMove : Byte;
  goalState: Byte;
begin
  numUnitPropagation := 0;
  numContradiction := 0;
  numProbability := 0;
  done := false;
  moveQueue.EmptyList();
  while not(done) do
  begin
    goalState := msi.GetGoalState();
    if((goalState = MSI_GAMEWON) or (goalState = MSI_GAMELOST)) Then
    begin
      Sleep(10);
      goalState := msi.GetGoalState();
      if((goalState = MSI_GAMEWON) or (goalState = MSI_GAMELOST)) Then
         break;
    end;
    lastMove := GetMove();
    case(lastMove) of
      MSAI_MOVE_UNITPROPAGATION : if(numUnitPropagation > 0)Then
                                    numUnitPropagation := numUnitPropagation +moveQueue.numMoves;
      MSAI_MOVE_CONTRADICTION   : if(numContradiction > 0) Then
                                    numContradiction  := numContradiction+moveQueue.numMoves;
      MSAI_MOVE_PROBABILITY     : if(numProbability > 0)   Then
                                    numProbability  := numProbability +moveQueue.numMoves;
      else output(Format('lastMove = %u',[lastMove]));
    end;
    if(not(stopOnProbability) or (lastMove <> MSAI_MOVE_PROBABILITY)) Then
      ExecuteQueue()
    else
      done := true;
  end;
  Result := goalState;
end;

Function TMinesweeperAI.ExecuteQueue() : Byte;
var
  x, y : Byte;
  rightClick : Boolean;
begin
  if(moveQueue.numMoves > 0) Then
  begin
    while(moveQueue.Dequeue(x,y,rightClick)) do
      msi.SimulateClick(x,y,rightClick);
    if(verbose[1]) Then
       output('Queue executed!');
    Result := MSAI_SUCCESS;
    Exit;
  end
  else if(verbose[1]) Then
    output('Queue is empty - cannot execute!');
  Result := Ord(MSAI_ERROR);
end;

Function TMinesweeperAI.GetMoveProbability() : Byte;
var
  probability: double ;
  x,y : Byte;
  lowest : double ;
  uncovered : Byte;
  li : TLargeInteger;
  move : PMoveNode;
  temp : MoveNode ;
  choice : SmallInt;
  a :integer;
begin
  lowest := 1.0;
  for x := 0 to pred(columns) do
    for y := 0 to pred(rows) do
    begin
      if(GetBoard(x,y) = MSI_SQUARE_BLANK) Then
      begin
	uncovered := GetNearUncovered(x, y);
	if (uncovered = 0) then
	  probability := numMinesLeft / numBlanks
       	else
	   probability := GetProbability(x,y);

	if(probability < lowest) then
	begin
	  moveQueue.EmptyList();
	  moveQueue.Enqueue(x, y);
	  lowest := probability;
	  if(verbose[2]) Then
	    output(Format('New lowest probability (%u,%u)=%g',[x, y, probability]));
	end
	else if(probability = lowest) Then
	   moveQueue.Enqueue(x,y);
	if(verbose[3]) Then
	   output(Format('Square %u,%u has probability %g (%u uncoverd)',[x, y, probability, uncovered]));
    end;
  end;

  if(moveQueue.numMoves > 1) Then
  begin
    Randomize;
    a := columns*rows;
    choice := Random(a);
    move := moveQueue.GetIndex(choice);
    Result := MSI_ERROR;
    if assigned(move) then
    begin
      if(verbose[1]) Then
        output(format('     ... Added move: %u, %u for uncovering to queue (probability of %g)',[move.x, move.y, lowest]));
      temp.x := move.x;
      temp.y := move.y;
      temp.rightClick := false;
      moveQueue.EmptyList();
      moveQueue.Enqueue(temp.x, temp.y, temp.rightClick);
    end;
    Result := MSAI_MOVE_PROBABILITY;
  end;
end;

Function TMinesweeperAI.GetProbability(col,row : Byte) : double;
var
  probability : Double;
  i, j: Shortint;
  numUsed : Byte;
  bombsLeft : Byte;
  blanks : Byte;
  symbol : Char;
begin
  probability := 0.0;
  i := 0;
  j := 0;
  numUsed := 0;

  for i := -1 to 1 do
    for j := -1 to 1 do
    begin
      symbol := GetBoard(col+i, row+j);
      if(((i <> 0)or (j <> 0)) and (ord(symbol) > 0) and (ord(symbol) <= 8)) Then
      begin
	GetNear(col+i, row+j, blanks, bombsLeft);
	bombsLeft := ord(symbol) - bombsLeft;
        if  blanks <> 0  Then
          probability := probability +(bombsLeft / blanks)
         else
          probability := 0.0 ;
	inc(numUsed);
      end;
    end;
  if(probability = 0.0) Then
    result :=  1.0
  else
    result := (probability / numUsed);
end;

Function TMinesweeperAI.Map2Dto1D(column,row : Byte) : SmallInt;
begin
  Result := row*columns+column;
end;

Function TMinesweeperAI.GetNear(col,row : Byte; Var blanks, flags : Byte) : Byte;
Var
  i, j : Shortint;
  symbol : Char;
begin
  blanks := 0;
  flags  := 0;
  for i := -1  to 1 do
    for j := -1 to  1 do
      if((i <> 0) or (j <> 0)) Then
      begin
        symbol := GetBoard(col+i, row+j);
	Case(symbol) of
	  MSI_SQUARE_BLANK : inc(blanks);
	  MSI_SQUARE_FLAG  : inc(flags);
	else
          break;
	end;
      end;
  Result := MSAI_SUCCESS;
end;

Function TMinesweeperAI.GetNearUncovered(col,row : Byte) : Byte;
var
  numUncovered : Byte;
  i, j : Shortint;
  symbol : Char;
begin
  numUncovered := 0;
  for i := -1 to 1 do
    for j := -1 to 1 do
      if((i <> 0) or (j <> 0)) Then
      begin
	symbol := GetBoard(col+i, row+j);
	if((ord(symbol) >= 0) and (ord(symbol) <= 8)) Then
	  Inc(numUncovered);
      end;
  result := numUncovered;
end;

Function TMinesweeperAI.QueueSurrounding(col,row : Byte;flag : Boolean) : Byte;
var
  i, j : ShortInt;
  numMoves : Integer;
begin
  numMoves := 0;
  for i := -1 to 1 do
    for j := -1 to 1 do    // if square is blank and a neighbour ...
      if (( (i <> 0) or(j <> 0)) and (GetBoard(col+i, row+j) = ' ')) Then
      begin
	if(moveQueue.Enqueue(col+i, row+j, flag) and verbose[2]) Then
	begin
	  inc(numMoves);
	  output(format('     Added move: (%u, %u) to %s',[col+i, row+j,
            iif(flag,'mark as mine','uncover')]));
	end;
      end;
  result := numMoves;
end;

end.
